/* /////////////////////////////////////////////////////////////////////////////////////////////////
Life#IdlePoser by MacGruber.
State-based idle animation system with anchoring.
https://www.patreon.com/MacGruber_Laboratory

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

#if !VAM_GT_1_20
	#error IdlePoser requires VaM 1.20 or newer!
#endif

using UnityEngine;
using UnityEngine.Events;
using System;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
	public partial class IdlePoser : MVRScript
	{
		private const int MAX_STATES = 4;
		private static readonly int[] DISTANCE_SAMPLES = new int[] { 0, 0, 0, 11, 20};

		private const int STATETYPE_REGULARSTATE = 0;
		private const int STATETYPE_CONTROLPOINT = 1;
		private const int STATETYPE_INTERMEDIATE = 2;
		private const int NUM_STATETYPES = 3;
		
		private const float DEFAULT_TRANSITION_DURATION = 0.5f;
		private const float DEFAULT_BLEND_DURATION = 0.2f;
		private const float DEFAULT_EASEIN_DURATION = 1.0f;
		private const float DEFAULT_EASEOUT_DURATION = 1.0f;
		private const float DEFAULT_PROBABILITY = 0.5f;
		private const float DEFAULT_WAIT_DURATION_MIN = 2.0f;
		private const float DEFAULT_WAIT_DURATION_MAX = 4.0f;
		private const float DEFAULT_ANCHOR_BLEND_RATIO = 0.5f;
		private const float DEFAULT_ANCHOR_DAMPING_TIME = 0.2f;

		private Dictionary<string, State> myStates = new Dictionary<string, State>();
		private List<ControlCapture> myControlCaptures = new List<ControlCapture>();
		private List<MorphCapture> myMorphCaptures = new List<MorphCapture>();
		private List<State> myTransition = new List<State>(8);
		private List<State> myCurrentTransition = new List<State>(MAX_STATES);
		private List<TriggerActionDiscrete> myTriggerActionsNeedingUpdate = new List<TriggerActionDiscrete>();
		private State myCurrentState;
		private State myNextState;
		private State myBlendState = State.CreateBlendState();

		private float myDuration = 1.0f;
		private float myClock = 0.0f;
		private uint myStateMask = 0;
		private bool myStateMaskChanged = false;
		private bool myNoValidTransition = false;
		private bool myPlayMode = false;
		private bool myPaused = false;
		private bool myNeedRefresh = false;
		private bool myWasLoading = true;

		private JSONStorableString mySwitchState;
		private JSONStorableString mySetStateMask;
		private JSONStorableString myPartialStateMask;

		public override void Init()
		{
			myWasLoading = true;
			myClock = 0.0f;

			ControlCapture lHand = new ControlCapture(this, "lHandControl");
			if (lHand.IsValid())
				myControlCaptures.Add(lHand);
			ControlCapture rHand = new ControlCapture(this, "rHandControl");
			if (rHand.IsValid())
				myControlCaptures.Add(rHand);
			ControlCapture control = new ControlCapture(this, "control");
			if (myControlCaptures.Count == 0 && control.IsValid())
				myControlCaptures.Add(control);	
			

			InitUI();


			mySwitchState = new JSONStorableString("SwitchState", "", SwitchStateAction);
			mySwitchState.isStorable = mySwitchState.isRestorable = false;
			RegisterString(mySwitchState);

			mySetStateMask = new JSONStorableString("SetStateMask", "", SetStateMaskAction);
			mySetStateMask.isStorable = mySetStateMask.isRestorable = false;
			RegisterString(mySetStateMask);
			
			myPartialStateMask = new JSONStorableString("PartialStateMask", "", PartialStateMaskAction);
			myPartialStateMask.isStorable = myPartialStateMask.isRestorable = false;
			RegisterString(myPartialStateMask);

			Utils.SetupAction(this, "TriggerSync", TriggerSyncAction);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			OnDestroyUI();

			foreach (var s in myStates)
			{
				State state = s.Value;
				state.EnterBeginTrigger.Remove();
				state.EnterEndTrigger.Remove();
				state.ExitBeginTrigger.Remove();
				state.ExitEndTrigger.Remove();
			}
		}

		private State CreateState(string name)
		{
			State s = new State(this, name) {
				myWaitDurationMin = DEFAULT_WAIT_DURATION_MIN,
				myWaitDurationMax = DEFAULT_WAIT_DURATION_MAX,
				myTransitionDuration = DEFAULT_TRANSITION_DURATION,
				myStateType = STATETYPE_REGULARSTATE
			};
			CaptureState(s);
			myStates[name] = s;
			return s;
		}

		private void SetState(State state)
		{			
			myNoValidTransition = false;
			myCurrentState = state;
			myNextState = null;
			myClock = 0.0f;
			if (state.myWaitInfiniteDuration)
				myDuration = float.MaxValue;
			else
				myDuration = UnityEngine.Random.Range(state.myWaitDurationMin, state.myWaitDurationMax);
			myCurrentTransition.Clear();
		}

		private void SetTransition(float duration = -1.0f)
		{
			float d = 0.0f;
			int entryCount = Mathf.Min(myTransition.Count, MAX_STATES);
			for (int i=0; i<entryCount; ++i)
			{
				d += myTransition[i].myTransitionDuration;
				if (i > 0 && !myTransition[i].IsControlPoint)
				{
					entryCount = i+1;
					break;
				}
			}

			myNoValidTransition = false;
			myCurrentTransition.Clear();
			for (int i=0; i<entryCount; ++i)
				myCurrentTransition.Add(myTransition[i]);

			myClock = 0.0f;
			myDuration = (duration < 0) ? d : duration;
			myDuration = Mathf.Max(myDuration, 0.001f);
			myNextState = myTransition[entryCount-1];
			for (int i=0; i<myControlCaptures.Count; ++i)
				myControlCaptures[i].SetTransition(myTransition, entryCount);
			for (int i=0; i<myMorphCaptures.Count; ++i)
				myMorphCaptures[i].SetTransition(myTransition, entryCount);

			if (myTransition.Count == entryCount)
				myTransition.Clear();
			else
				myTransition.RemoveRange(0, entryCount-1);

			if (myCurrentState.ExitBeginTrigger != null)
				myCurrentState.ExitBeginTrigger.Trigger(myTriggerActionsNeedingUpdate);
			if (myNextState.EnterBeginTrigger != null)
				myNextState.EnterBeginTrigger.Trigger(myTriggerActionsNeedingUpdate);
		}

		private void SetRandomTransition()
		{
			List<State> states = new List<State>(16);
			myTransition.Clear();
			myTransition.Add(myCurrentState);
			myNextState = null;
			GatherStates(1, states);

			int i;
			float sum = 0.0f;
			for (i=0; i<states.Count; ++i)
				sum += states[i].myProbability;
			if (sum == 0.0f)
			{
				myTransition.Clear();
				myNoValidTransition = true;
			}
			else
			{
				float threshold = UnityEngine.Random.Range(0.0f, sum);
				sum = 0.0f;
				for (i=0; i<states.Count-1; ++i)
				{
					sum += states[i].myProbability;
					if (threshold <= sum)
						break;
				}
				GatherTransition(1, i, 0);
				SetTransition();
			}
		}

		private void SetBlendTransition(State state, bool debug = false)
		{
			myTransition.Clear();
			myTransition.Add(myCurrentState);
			if (myCurrentState != null)
			{
				List<State> states = new List<State>(16);
				myNextState = null;
				GatherStates(1, states);
				List<int> indices = new List<int>(4);
				for (int i=0; i<states.Count; ++i)
				{
					if (states[i] == state)
						indices.Add(i);
				}
				if (indices.Count == 0)
				{
					states.Clear();
					myNextState = state;
					GatherStates(1, states);
					for (int i=0; i<states.Count; ++i)
					{
						if (states[i] == state)
							indices.Add(i);
					}
				}
				if (indices.Count > 0)
				{
					int selected = UnityEngine.Random.Range(0, indices.Count);
					GatherTransition(1, indices[selected], 0);
				}
			}

			if (myCurrentState == null || debug)
			{
				CaptureState(myBlendState);
				myTransition[0] = myBlendState;				
				if (myCurrentState != null && myTransition.Count > 1)
				{
					myBlendState.myTransitionDuration = myCurrentState.myTransitionDuration;
					myBlendState.myEaseInDuration = myCurrentState.myEaseInDuration;
					myBlendState.myEaseOutDuration = myCurrentState.myEaseOutDuration;	
				}
				else
				{
					myBlendState.myTransitionDuration = DEFAULT_BLEND_DURATION;
					myBlendState.myEaseInDuration = DEFAULT_EASEIN_DURATION;
					myBlendState.myEaseOutDuration = DEFAULT_EASEOUT_DURATION;
				}
				myBlendState.AssignOutTriggers(myCurrentState);
				SetState(myBlendState);
			}
			
			if (myTransition.Count == 1) // Did not find transition....fake one
			{
				myTransition.Add(state);
			}
			
			SetTransition();
		}

		private void GatherStates(int transitionLength, List<State> states)
		{
			State source = myTransition[0];
			State current = myTransition[myTransition.Count-1];
			for (int i=0; i<current.myTransitions.Count; ++i)
			{
				State next = current.myTransitions[i];
				if (myTransition.Contains(next))
					continue;

				if (next.IsRegularState || next == myNextState)
				{
					if (DoAcceptRegularState(source, next))
						states.Add(next);
				}
				else if (next.IsControlPoint)
				{
					if (transitionLength >= MAX_STATES-1)
						continue;
					myTransition.Add(next);
					GatherStates(transitionLength+1, states);
					myTransition.RemoveAt(myTransition.Count-1);
				}
				else // next.IsIntermediate
				{
					myTransition.Add(next);
					GatherStates(1, states);
					myTransition.RemoveAt(myTransition.Count-1);
				}
			}
		}

		private int GatherTransition(int transitionLength, int selected, int index)
		{
			State source = myTransition[0];
			State current = myTransition[myTransition.Count-1];
			for (int i=0; i<current.myTransitions.Count; ++i)
			{
				State next = current.myTransitions[i];
				if (myTransition.Contains(next))
					continue;

				if (next.IsRegularState || next == myNextState)
				{
					if (!DoAcceptRegularState(source, next))
					{
						continue;
					}
					else if (index == selected)
					{
						myTransition.Add(next);
						return -1;
					}
					else
					{
						++index;
					}
				}
				else if (next.IsControlPoint)
				{
					if (transitionLength >= MAX_STATES-1)
						continue;
					myTransition.Add(next);
					index = GatherTransition(transitionLength+1, selected, index);
					if (index == -1)
						return -1;
					myTransition.RemoveAt(myTransition.Count-1);
				}
				else // next.IsIntermediate
				{
					myTransition.Add(next);
					index = GatherTransition(1, selected, index);
					if (index == -1)
						return -1;
					myTransition.RemoveAt(myTransition.Count-1);
				}
			}
			return index;
		}
		
		private bool DoAcceptRegularState(State source, State next, bool debugMode = false)
		{
			if (next == myNextState && !debugMode)
				return true;
			if (next.myProbability < 0.01f)
				return false;
			if (myStateMask != 0 && (myStateMask & next.StateMask) == 0 && !debugMode)
				return false;										
			if (source.myStateGroup == 0 || source.myStateGroup != next.myStateGroup)
				return true;
			
			// in-group transition: source.myStateGroup == next.myStateGroup
			if (!source.myAllowInGroupTransition || !next.myAllowInGroupTransition)
				return false;	

			List<State> transition = debugMode ? myDebugTransition : myTransition;
			for (int t=1; t<transition.Count; ++t)
			{
				if (!transition[t].myAllowInGroupTransition)
					return false;
			}
			
			return true;
		}

		private void BlendToRandomState(float duration)
		{
			List<State> possible = new List<State>(myStates.Count);
			foreach (var state in myStates)
			{
				if (state.Value.IsRegularState)
					possible.Add(state.Value);
			}
			int idx = UnityEngine.Random.Range(0, possible.Count);

			CaptureState(myBlendState);
			myTransition.Clear();
			myTransition.Add(myBlendState);
			myTransition.Add(possible[idx]);

			myBlendState.AssignOutTriggers(myCurrentState);
			SetState(myBlendState);
			SetTransition(duration);
		}

		private void CaptureState(State state)
		{
			for (int i=0; i<myControlCaptures.Count; ++i)
				myControlCaptures[i].CaptureEntry(state);
			for (int i=0; i<myMorphCaptures.Count; ++i)
				myMorphCaptures[i].CaptureEntry(state);
		}
		
		private void SwitchStateAction(string v)
		{
			mySwitchState.valNoCallback = string.Empty;

			State state;
			if (myStates.TryGetValue(v, out state))
				SetBlendTransition(state);
			else
				SuperController.LogError("IdlePoser: Can't switch to unknown state '"+v+"'!");
		}

		private void SetStateMaskAction(string v)
		{
			mySetStateMask.valNoCallback = string.Empty;

			myStateMask = 0;
			bool invert = false;
			bool error = false;
			if (v.ToLower() != "clear")
			{
				for (int i=0; i<v.Length; ++i)
				{
					char c = v[i];
					if (i == 0 && c == '!')
						invert = true;
					else if (c >= 'A' && c <= 'L')
						myStateMask |= 1u << (c - 'A' + 1);
					else if (c >= 'a' && c <= 'l')
						myStateMask |= 1u << (c - 'a' + 1);
					else if (c == 'N' || c == 'n')
						myStateMask |= 1u << 0;
					else
						error = true;
				}
			}

			if (error)
			{
				SuperController.LogError("IdlePoser: SetStateMask set to invalid data: '"+v+"'");
				return;
			}

			if (myStateMask != 0 && invert)
				myStateMask = ~myStateMask;

			myStateMaskChanged = true;
			TryApplyStateMaskChange();
		}
		
		private void PartialStateMaskAction(string v)
		{
			myPartialStateMask.valNoCallback = string.Empty;

			uint stateMask = 0;
			bool invert = false;
			bool error = false;
			string[] masks = v.Split(new char[]{' ',';','|'}, 32, StringSplitOptions.RemoveEmptyEntries);
			for (int m=0; m<masks.Length; ++m)
			{
				stateMask = 0;
				invert = false;
				for (int i=0; i<masks[m].Length; ++i)
				{
					char c = masks[m][i];
					if (i == 0 && c == '!')
						invert = true;
					else if (c >= 'A' && c <= 'L')
						stateMask |= 1u << (c - 'A' + 1);
					else if (c >= 'a' && c <= 'l')
						stateMask |= 1u << (c - 'a' + 1);
					else if (c == 'N' || c == 'n')
						stateMask |= 1u << 0;
					else
						error = true;
				}

				if (error)
				{
					SuperController.LogError("IdlePoser: PartialStateMask set to invalid data: '"+v+"'");
					return;
				}

				if (stateMask == 0)
					continue;
					
				if (invert)
					myStateMask = myStateMask & ~stateMask;
				else
					myStateMask = myStateMask | stateMask;
			}

			myStateMaskChanged = true;
			TryApplyStateMaskChange();
		}
		
		private void TryApplyStateMaskChange()
		{
			if (myCurrentState == null || myNextState != null || !myCurrentState.IsRegularState)
				return;
							
			if (myClock >= myDuration && !myCurrentState.myWaitForSync && !myPaused)
			{
				myStateMaskChanged = false;
				SetRandomTransition();
			}
			else if (myStateMask != 0 && (myCurrentState.StateMask & myStateMask) == 0)
			{
				myStateMaskChanged = false;
				SetRandomTransition();
			}
		}

		private void TriggerSyncAction()
		{
			if (myCurrentState != null && myNextState == null
			 && myClock >= myDuration && !myCurrentState.myWaitInfiniteDuration
			 && myCurrentState.myWaitForSync && !myPaused)
			{
				if (myTransition.Count == 0)
					SetRandomTransition();
				else
					SetTransition();
			}
		}

		private void Update()
		{
			bool isLoading = SuperController.singleton.isLoading;
			bool isOrWasLoading = isLoading || myWasLoading;
			myWasLoading = isLoading;
			if (isOrWasLoading)
				return;

			if (myNeedRefresh)
			{
				UIRefreshMenu();
				myNeedRefresh = false;
			}
			DebugUpdateUI();

			for (int i=0; i<myTriggerActionsNeedingUpdate.Count; ++i)
				myTriggerActionsNeedingUpdate[i].Update();
			myTriggerActionsNeedingUpdate.RemoveAll(a => !a.timerActive);

			if (myCurrentState == null)
				return;

			bool paused = myPaused && myNextState == null && myTransition.Count == 0;
			if (!paused)
				myClock = Mathf.Min(myClock + Time.deltaTime, 100000.0f);

			if (myNextState != null)
			{
				float t = Smooth(myCurrentState.myEaseOutDuration, myNextState.myEaseInDuration, myDuration, myClock);
				for (int i=0; i<myControlCaptures.Count; ++i)
					myControlCaptures[i].UpdateTransition(t);
				for (int i=0; i<myMorphCaptures.Count; ++i)
					myMorphCaptures[i].UpdateTransition(t);
			}
			else if (!paused || myPlayMode)
			{
				for (int i=0; i<myControlCaptures.Count; ++i)
					myControlCaptures[i].UpdateState(myCurrentState);
			}		
			
			if (myClock >= myDuration)
			{
				if (myNextState != null)
				{
					State previousState = myCurrentState;	
					SetState(myNextState);
					if (myTransition.Count == 0)
						myMainState.valNoCallback = myCurrentState.myName;
					
					if (previousState.ExitEndTrigger != null)
						previousState.ExitEndTrigger.Trigger(myTriggerActionsNeedingUpdate);
					if (myCurrentState.EnterEndTrigger != null)
						myCurrentState.EnterEndTrigger.Trigger(myTriggerActionsNeedingUpdate);
				}
				else if (!paused && !myCurrentState.myWaitForSync && !myNoValidTransition)
				{
					if (myTransition.Count == 0)
						SetRandomTransition();
					else
						SetTransition();
				}
				else if (myStateMaskChanged)
				{
					TryApplyStateMaskChange();
				}
			}
			else if (myStateMaskChanged)
			{
				TryApplyStateMaskChange();
			}
		}

		private float Smooth(float a, float b, float d, float t)
		{
			d = Mathf.Max(d, 0.01f);
			t = Mathf.Clamp(t, 0.0f, d);			
			if (a+b>d)
			{
				float scale = d/(a+b);
				a *= scale;
				b *= scale;
			}			
			float n = d - 0.5f*(a+b);	
			float s = d - t;			

			// This is based on using the SmoothStep function (3x^2 - 2x^3) for velocity: https://en.wikipedia.org/wiki/Smoothstep
			// The result is a 3-piece curve consiting of a linear part in the middle and the integral of SmoothStep at both
			// ends. Additionally there is some scaling to connect the parts properly.
			// The resulting combined curve has smooth velocity and continuous acceleration/deceleration.
			float ta = t / a;
			float sb = s / b;
			if (t < a)
				return (a - 0.5f*t) * (ta*ta*ta/n);
			else if (s >= b)
				return (t - 0.5f*a) / n;
			else		
				return (0.5f*s - b) * (sb*sb*sb/n) + 1.0f;
			
			/*
			// Alternative using SmootherStep (6x^5 - 15x^4 + 10x^3) as a basis instead of SmoothStep
			
			if (t < a)
			{
				float t2 = t*t;
				float a2 = a*a;
				return ( t2/a2 - 3.0f*t/a + 2.5f ) * (t2*t2) / (n*a2*a);
			}
			else if (s >= b)
			{
				return (t - 0.5f*a) / n;
			}
			else
			{					
				float s2 = s*s;
				float b2 = b*b;
				return ( -s2/b2 + 3.0f*s/b - 2.5f ) * (s2*s2) / (n*b2*b) + 1.0f;
			}
			
			*/
		}

		private void OnAtomRename(string oldid, string newid)
		{
			foreach (var s in myStates)
			{
				State state = s.Value;
				state.EnterBeginTrigger.SyncAtomNames();
				state.EnterEndTrigger.SyncAtomNames();
				state.ExitBeginTrigger.SyncAtomNames();
				state.ExitEndTrigger.SyncAtomNames();
				
				foreach (var ce in state.myControlEntries)
				{
					ControlEntryAnchored entry = ce.Value;
					if (entry.myAnchorAAtom == oldid)
						entry.myAnchorAAtom = newid;
					if (entry.myAnchorBAtom == oldid)
						entry.myAnchorBAtom = newid;
				}
			}
			
			myNeedRefresh = true;
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if ((includePhysical && includeAppearance) || forceStore) // StoreType.Full
			{
				jc["idlepose"] = SavePose();
				needsStore = true;
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (restorePhysical && restoreAppearance) // StoreType.Full
			{
				if (jc.HasKey("idlepose"))
					LoadPose(jc["idlepose"].AsObject);
				myNeedRefresh = true;
			}
		}

		private JSONClass SavePose()
		{
			JSONClass jc = new JSONClass();

			// save info
			JSONClass info = new JSONClass();
			info["Format"] = "MacGruber.IdlePoser";
			info["Version"].AsInt = 9;
			string creatorName = UserPreferences.singleton.creatorName;
			if (string.IsNullOrEmpty(creatorName))
				creatorName = "Unknown";
			info["Author"] = creatorName;
			jc["Info"] = info;

			// save settings
			if (myCurrentState != null)
				jc["InitialState"] = myCurrentState.myName;
			jc["Paused"].AsBool = myPlayPaused.val;
			jc["StateMask"].AsInt = (int)myStateMask;
			jc["DefaultToWorldAnchor"].AsBool = myOptionsDefaultToWorldAnchor.val;

			// save captures
			if (myControlCaptures.Count > 0)
			{
				JSONArray cclist = new JSONArray();
				for (int i=0; i<myControlCaptures.Count; ++i)
				{
					ControlCapture cc = myControlCaptures[i];
					JSONClass ccclass = new JSONClass();
					ccclass["Name"] = cc.myName;
					ccclass["ApplyPos"].AsBool = cc.myApplyPosition;
					ccclass["ApplyRot"].AsBool = cc.myApplyRotation;
					cclist.Add("", ccclass);
				}
				jc["ControlCaptures"] = cclist;
			}
			if (myMorphCaptures.Count > 0)
			{
				JSONArray mclist = new JSONArray();
				for (int i=0; i<myMorphCaptures.Count; ++i)
				{
					MorphCapture mc = myMorphCaptures[i];
					JSONClass mcclass = new JSONClass();
					mcclass["UID"] = mc.myMorph.uid;
					mcclass["SID"] = mc.mySID;
					mcclass["Apply"].AsBool = mc.myApply;
					mclist.Add("", mcclass);
				}
				jc["MorphCaptures"] = mclist;
			}

			// save states
			JSONArray slist = new JSONArray();
			foreach (var s in myStates)
			{
				State state = s.Value;
				JSONClass st = new JSONClass();
				st["Name"] = state.myName;
				st["WaitInfiniteDuration"].AsBool = state.myWaitInfiniteDuration;
				st["WaitForSync"].AsBool = state.myWaitForSync;
				st["AllowInGroupT"].AsBool = state.myAllowInGroupTransition;
				st["WaitDurationMin"].AsFloat = state.myWaitDurationMin;
				st["WaitDurationMax"].AsFloat = state.myWaitDurationMax;
				st["TransitionDuration"].AsFloat = state.myTransitionDuration;
				st["EaseInDuration"].AsFloat = state.myEaseInDuration;
				st["EaseOutDuration"].AsFloat = state.myEaseOutDuration;
				st["Probability"].AsFloat = state.myProbability;
				st["StateType"].AsInt = state.myStateType;
				st["StateGroup"].AsInt = state.myStateGroup;

				JSONArray tlist = new JSONArray();
				for (int i=0; i<state.myTransitions.Count; ++i)
					tlist.Add("", state.myTransitions[i].myName);
				st["Transitions"] = tlist;

				if (state.myControlEntries.Count > 0)
				{
					JSONClass celist = new JSONClass();
					foreach (var e in state.myControlEntries)
					{
						ControlEntryAnchored ce = e.Value;
						JSONClass ceclass = new JSONClass();
						ceclass["PX"].AsFloat = ce.myAnchorOffset.myPosition.x;
						ceclass["PY"].AsFloat = ce.myAnchorOffset.myPosition.y;
						ceclass["PZ"].AsFloat = ce.myAnchorOffset.myPosition.z;
						Vector3 rotation = ce.myAnchorOffset.myRotation.eulerAngles;
						ceclass["RX"].AsFloat = rotation.x;
						ceclass["RY"].AsFloat = rotation.y;
						ceclass["RZ"].AsFloat = rotation.z;
						ceclass["AnchorMode"].AsInt = ce.myAnchorMode;
						if (ce.myAnchorMode >= ControlEntryAnchored.ANCHORMODE_SINGLE)
						{
							ceclass["DampingTime"].AsFloat = ce.myDampingTime;
							ceclass["AnchorAAtom"] = ce.myAnchorAAtom;
							ceclass["AnchorAControl"] = ce.myAnchorAControl;
						}
						if (ce.myAnchorMode == ControlEntryAnchored.ANCHORMODE_BLEND)
						{
							ceclass["AnchorBAtom"] = ce.myAnchorBAtom;
							ceclass["AnchorBControl"] = ce.myAnchorBControl;
							ceclass["BlendRatio"].AsFloat = ce.myBlendRatio;
						}
						celist[e.Key.myName] = ceclass;
					}
					st["ControlEntries"] = celist;
				}

				if (state.myMorphEntries.Count > 0)
				{
					JSONClass melist = new JSONClass();
					foreach (var e in state.myMorphEntries)
					{
						melist[e.Key.mySID].AsFloat = e.Value;
					}
					st["MorphEntries"] = melist;
				}

				st[state.EnterBeginTrigger.Name] = state.EnterBeginTrigger.GetJSON(base.subScenePrefix);
				st[state.EnterEndTrigger.Name] = state.EnterEndTrigger.GetJSON(base.subScenePrefix);
				st[state.ExitBeginTrigger.Name] = state.ExitBeginTrigger.GetJSON(base.subScenePrefix);
				st[state.ExitEndTrigger.Name] = state.ExitEndTrigger.GetJSON(base.subScenePrefix);

				slist.Add("", st);
			}
			jc["States"] = slist;

			return jc;
		}

		private void LoadPose(JSONClass jc)
		{
			// reset
			foreach (var s in myStates)
			{
				State state = s.Value;
				state.EnterBeginTrigger.Remove();
				state.EnterEndTrigger.Remove();
				state.ExitBeginTrigger.Remove();
				state.ExitEndTrigger.Remove();
			}

			myControlCaptures.Clear();
			myMorphCaptures.Clear();
			myStates.Clear();
			myCurrentState = null;
			myNextState = null;
			myBlendState.myControlEntries.Clear();
			myBlendState.myMorphEntries.Clear();
			myClock = 0.0f;

			// load info
			int version = jc["Info"].AsObject["Version"].AsInt;

			// load captures
			if (jc.HasKey("ControlCaptures"))
			{
				JSONArray cclist = jc["ControlCaptures"].AsArray;
				for (int i=0; i<cclist.Count; ++i)
				{
					ControlCapture cc;
					if (version <= 2)
					{
						// handling legacy
						cc = new ControlCapture(this, cclist[i].Value);
					}
					else
					{
						JSONClass ccclass = cclist[i].AsObject;
						cc = new ControlCapture(this, ccclass["Name"]);
						cc.myApplyPosition = ccclass["ApplyPos"].AsBool;
						cc.myApplyRotation = ccclass["ApplyRot"].AsBool;
					}

					if (cc.IsValid())
						myControlCaptures.Add(cc);
				}
			}
			if (jc.HasKey("MorphCaptures"))
			{
				JSONArray mclist = jc["MorphCaptures"].AsArray;
				DAZCharacterSelector geometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
				for (int i=0; i<mclist.Count; ++i)
				{
					MorphCapture mc;
					if (version >= 9)
					{
						JSONClass mcclass = mclist[i].AsObject;
						string uid = mcclass["UID"];
						if (uid.EndsWith(".vmi")) // handle custom morphs, resolve VAR packages
							uid = SuperController.singleton.NormalizeLoadPath(uid);
						string sid = mcclass["SID"];
						mc = new MorphCapture(geometry, uid, sid);
						mc.myApply = mcclass["Apply"].AsBool;
					}
					else if (version == 8)
					{
						// handling legacy
						JSONClass mcclass = mclist[i].AsObject;
						mc = new MorphCapture(this, geometry, mcclass["UID"], mcclass["IsFemale"].AsBool);
						mc.myApply = mcclass["Apply"].AsBool;
					}
					else if (version >= 3 && version <= 7)
					{
						// handling legacy
						JSONClass mcclass = mclist[i].AsObject;
						mc = new MorphCapture(this, geometry, mcclass["Name"]);
						mc.myApply = mcclass["Apply"].AsBool;
					}
					else // (version <= 2)
					{
						// handling legacy
						mc = new MorphCapture(this, geometry, mclist[i].Value);
					}

					if (mc.IsValid())
					{
						myMorphCaptures.Add(mc);
					}
				}
			}

			// load states
			JSONArray slist = jc["States"].AsArray;
			for (int i=0; i<slist.Count; ++i)
			{
				// load state
				JSONClass st = slist[i].AsObject;
				State state;
				if (version == 1)
				{
					// handling legacy
					state = new State(this, st["Name"]) {
						myWaitDurationMin = st["WaitDurationMin"].AsFloat,
						myWaitDurationMax = st["WaitDurationMax"].AsFloat,
						myTransitionDuration = st["TransitionDuration"].AsFloat,
						myProbability = DEFAULT_PROBABILITY,
						myStateType = st["IsTransition"].AsBool ? STATETYPE_CONTROLPOINT : STATETYPE_REGULARSTATE,
						myStateGroup = 0
					};
				}
				else
				{
					state = new State(this, st["Name"]) {
						myWaitInfiniteDuration = st["WaitInfiniteDuration"].AsBool,
						myWaitForSync = st["WaitForSync"].AsBool,
						myAllowInGroupTransition = st["AllowInGroupT"].AsBool,
						myWaitDurationMin = st["WaitDurationMin"].AsFloat,
						myWaitDurationMax = st["WaitDurationMax"].AsFloat,
						myTransitionDuration = st["TransitionDuration"].AsFloat,
						myEaseInDuration = st.HasKey("EaseInDuration") ? st["EaseInDuration"].AsFloat : DEFAULT_EASEIN_DURATION,
						myEaseOutDuration = st.HasKey("EaseOutDuration") ? st["EaseOutDuration"].AsFloat : DEFAULT_EASEOUT_DURATION,
						myProbability = st["Probability"].AsFloat,
						myStateType = st["StateType"].AsInt,
						myStateGroup = st["StateGroup"].AsInt
					};
				}

				state.EnterBeginTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);
				state.EnterEndTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);
				state.ExitBeginTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);
				state.ExitEndTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);


				if (myStates.ContainsKey(state.myName))
					continue;
				myStates[state.myName] = state;

				// load control captures
				if (myControlCaptures.Count > 0)
				{
					JSONClass celist = st["ControlEntries"].AsObject;
					foreach (string ccname in celist.Keys)
					{
						ControlCapture cc = myControlCaptures.Find(x => x.myName == ccname);
						if (cc == null)
							continue;

						JSONClass ceclass = celist[ccname].AsObject;
						ControlEntryAnchored ce = new ControlEntryAnchored(this, ccname);
						ce.myAnchorOffset.myPosition.x = ceclass["PX"].AsFloat;
						ce.myAnchorOffset.myPosition.y = ceclass["PY"].AsFloat;
						ce.myAnchorOffset.myPosition.z = ceclass["PZ"].AsFloat;
						Vector3 rotation;
						rotation.x = ceclass["RX"].AsFloat;
						rotation.y = ceclass["RY"].AsFloat;
						rotation.z = ceclass["RZ"].AsFloat;
						ce.myAnchorOffset.myRotation.eulerAngles = rotation;
						ce.myAnchorMode = ceclass["AnchorMode"].AsInt;
						if (ce.myAnchorMode >= ControlEntryAnchored.ANCHORMODE_SINGLE)
						{
							ce.myDampingTime = ceclass["DampingTime"].AsFloat;
							ce.myAnchorAAtom = ceclass["AnchorAAtom"].Value;
							ce.myAnchorAControl = ceclass["AnchorAControl"].Value;
							
							if (ce.myAnchorAAtom == "[Self]") // legacy
								ce.myAnchorAAtom = containingAtom.uid;
						}
						if (ce.myAnchorMode == ControlEntryAnchored.ANCHORMODE_BLEND)
						{
							ce.myAnchorBAtom = ceclass["AnchorBAtom"].Value;
							ce.myAnchorBControl = ceclass["AnchorBControl"].Value;
							ce.myBlendRatio = ceclass["BlendRatio"].AsFloat;
							
							if (ce.myAnchorBAtom == "[Self]") // legacy
								ce.myAnchorBAtom = containingAtom.uid;
						}
						ce.Initialize();

						state.myControlEntries.Add(cc, ce);
					}
					for (int j=0; j<myControlCaptures.Count; ++j)
					{
						if (!state.myControlEntries.ContainsKey(myControlCaptures[j]))
							myControlCaptures[j].CaptureEntry(state);
					}
				}

				// load morph captures
				if (myMorphCaptures.Count > 0)
				{
					JSONClass melist = st["MorphEntries"].AsObject;
					foreach (string key in melist.Keys)
					{
						MorphCapture mc = null;
						if (version >= 9)
						{
							mc = myMorphCaptures.Find(x => x.mySID == key);
						}
						else if (version == 8)
						{
							// legacy handling of old qualifiedName
							string uid;
							DAZCharacterSelector.Gender gender;
							if (key.EndsWith("#Female"))
							{
								uid = key.Substring(0, key.Length-7);
								gender = DAZCharacterSelector.Gender.Female;
							}
							else if (key.EndsWith("#Male"))
							{
								uid = key.Substring(0, key.Length-5);
								gender = DAZCharacterSelector.Gender.Male;
							}
							else
							{
								continue;
							}
								
							mc = myMorphCaptures.Find(x => x.myGender == gender && x.myMorph.uid == uid);
						}
						else // version <= 7
						{						
							// legacy handling of old qualifiedName where the order was reversed
							string uid;
							DAZCharacterSelector.Gender gender;
							if (key.StartsWith("Female#"))
							{
								uid = key.Substring(7);
								gender = DAZCharacterSelector.Gender.Female;
							}
							else if (key.StartsWith("Male#"))
							{
								uid = key.Substring(5);
								gender = DAZCharacterSelector.Gender.Male;
							}
							else
							{
								continue;
							}
								
							mc = myMorphCaptures.Find(x => x.myGender == gender && x.myMorph.uid == uid);
						}
						
						if (mc == null)
						{
							continue;
						}
						float me = melist[key].AsFloat;
						state.myMorphEntries.Add(mc, me);
					}
					for (int j=0; j<myMorphCaptures.Count; ++j)
					{
						if (!state.myMorphEntries.ContainsKey(myMorphCaptures[j]))
							myMorphCaptures[j].CaptureEntry(state);
					}
				}
			}

			// load transitions
			for (int i=0; i<slist.Count; ++i)
			{
				JSONClass st = slist[i].AsObject;
				State source;
				if (!myStates.TryGetValue(st["Name"], out source))
					continue;

				JSONArray tlist = st["Transitions"].AsArray;
				for (int j=0; j<tlist.Count; ++j)
				{
					State target;
					if (myStates.TryGetValue(tlist[j].Value, out target))
						source.myTransitions.Add(target);
				}
			}

			// load settings
			myPlayPaused.valNoCallback = jc.HasKey("Paused") && jc["Paused"].AsBool;
			myPlayPaused.setCallbackFunction(myPlayPaused.val);

			int stateMask = jc["StateMask"].AsInt;
			myStateMask = (uint)stateMask;
			
			myOptionsDefaultToWorldAnchor.val = jc.HasKey("DefaultToWorldAnchor") && jc["DefaultToWorldAnchor"].AsBool;

			// blend to initial state
			if (jc.HasKey("InitialState"))
			{
				State initial;
				if (myStates.TryGetValue(jc["InitialState"].Value, out initial))
				{
					SetState(initial);
					myMainState.valNoCallback = initial.myName;
				}
			}
		}


		// =======================================================================================


		private class State
		{
			public string myName;
			public float myWaitDurationMin;
			public float myWaitDurationMax;
			public float myTransitionDuration;
			public float myEaseInDuration = DEFAULT_EASEIN_DURATION;
			public float myEaseOutDuration = DEFAULT_EASEOUT_DURATION;
			public float myProbability = DEFAULT_PROBABILITY;
			public int myStateType;
			public int myStateGroup;
			public bool myWaitInfiniteDuration = false;
			public bool myWaitForSync = false;
			public bool myAllowInGroupTransition = false;
			public uint myDebugIndex = 0;
			public Dictionary<ControlCapture, ControlEntryAnchored> myControlEntries = new Dictionary<ControlCapture, ControlEntryAnchored>();
			public Dictionary<MorphCapture, float> myMorphEntries = new Dictionary<MorphCapture, float>();
			public List<State> myTransitions = new List<State>();
			public EventTrigger EnterBeginTrigger;
			public EventTrigger EnterEndTrigger;
			public EventTrigger ExitBeginTrigger;
			public EventTrigger ExitEndTrigger;

			public bool IsRegularState { get { return myStateType == STATETYPE_REGULARSTATE; } }
			public bool IsControlPoint { get { return myStateType == STATETYPE_CONTROLPOINT; } }
			public bool IsIntermediate { get { return myStateType == STATETYPE_INTERMEDIATE; } }
			public uint StateMask { get { return 1u << myStateGroup; } }

			private State(string name)
			{
				myName = name;
				// do NOT init event triggers
			}

			public State(MVRScript script, string name)
			{
				myName = name;
				EnterBeginTrigger = new EventTrigger(script, "OnEnterBegin", name);
				EnterEndTrigger = new EventTrigger(script, "OnEnterEnd", name);
				ExitBeginTrigger = new EventTrigger(script, "OnExitBegin", name);
				ExitEndTrigger = new EventTrigger(script, "OnExitEnd", name);
			}
			
			public State(string name, State source)
			{
				myName = name;
				myWaitDurationMin = source.myWaitDurationMin;
				myWaitDurationMax = source.myWaitDurationMax;
				myTransitionDuration = source.myTransitionDuration;
				myEaseInDuration = source.myEaseInDuration;
				myEaseOutDuration = source.myEaseOutDuration;
				myProbability = source.myProbability;
				myStateType = source.myStateType;
				myStateGroup = source.myStateGroup;
				myWaitInfiniteDuration = source.myWaitInfiniteDuration;
				myWaitForSync = source.myWaitForSync;
				myAllowInGroupTransition = source.myAllowInGroupTransition;
				EnterBeginTrigger = new EventTrigger(source.EnterBeginTrigger);
				EnterEndTrigger = new EventTrigger(source.EnterEndTrigger);
				ExitBeginTrigger = new EventTrigger(source.ExitBeginTrigger);
				ExitEndTrigger = new EventTrigger(source.ExitEndTrigger);
			}

			public static State CreateBlendState()
			{
				return new State("BlendState") {
					myWaitDurationMin = 0.0f,
					myWaitDurationMax = 0.0f,
					myTransitionDuration = 0.2f,
					myStateType = STATETYPE_CONTROLPOINT
				};
			}

			public void AssignOutTriggers(State other)
			{
				ExitBeginTrigger = other?.ExitBeginTrigger;
				ExitEndTrigger = other?.ExitEndTrigger;
			}
		}		

		private class ControlCapture
		{
			public string myName;
			private IdlePoser myPlugin;
			private Transform myTransform;
			private ControlEntryAnchored[] myTransition = new ControlEntryAnchored[MAX_STATES];
			private int myEntryCount = 0;
			public bool myApplyPosition = true;
			public bool myApplyRotation = true;

			private static Quaternion[] ourTempQuaternions = new Quaternion[MAX_STATES-1];
			private static float[] ourTempDistances = new float[DISTANCE_SAMPLES[DISTANCE_SAMPLES.Length-1] + 2];

			public ControlCapture(IdlePoser plugin, string control)
			{
				myPlugin = plugin;
				myName = control;
				FreeControllerV3 controller = plugin.containingAtom.GetStorableByID(control) as FreeControllerV3;
				if (controller != null)
					myTransform = controller.transform;
			}

			public void CaptureEntry(State state)
			{
				ControlEntryAnchored entry;
				if (!state.myControlEntries.TryGetValue(this, out entry))
				{
					entry = new ControlEntryAnchored(myPlugin, myName);
					entry.Initialize();
					state.myControlEntries[this] = entry;
				}
				entry.Capture(myTransform.position, myTransform.rotation);
			}

			public void SetTransition(List<State> states, int entryCount)
			{
				myEntryCount = entryCount;
				for (int i=0; i<myEntryCount; ++i)
				{
					if (!states[i].myControlEntries.TryGetValue(this, out myTransition[i]))
					{
						CaptureEntry(states[i]);
						myTransition[i] = states[i].myControlEntries[this];
					}
				}
			}

			public void UpdateTransition(float t)
			{
				for (int i=0; i<myEntryCount; ++i)
					myTransition[i].Update();

				//t = ArcLengthParametrization(t);

				if (myApplyPosition)
				{
					switch (myEntryCount)
					{
						case 4:	myTransform.position = EvalBezierCubicPosition(t);          break;
						case 3: myTransform.position = EvalBezierQuadraticPosition(t);      break;
						case 2: myTransform.position = EvalBezierLinearPosition(t);         break;
						default: myTransform.position = myTransition[0].myEntry.myPosition; break;
					}
				}
				if (myApplyRotation)
				{
					switch (myEntryCount)
					{
						case 4: myTransform.rotation = EvalBezierCubicRotation(t);          break;
						case 3: myTransform.rotation = EvalBezierQuadraticRotation(t);      break;
						case 2: myTransform.rotation = EvalBezierLinearRotation(t);         break;
						default: myTransform.rotation = myTransition[0].myEntry.myRotation; break;
					}
				}
			}

			private float ArcLengthParametrization(float t)
			{
				if (myEntryCount <= 2 || myEntryCount > 4)
					return t;

				int numSamples = DISTANCE_SAMPLES[myEntryCount];
				float numLines = (float)(numSamples+1);
				float distance = 0.0f;
				Vector3 previous = myTransition[0].myEntry.myPosition;
				ourTempDistances[0] = 0.0f;

				if (myEntryCount == 3)
				{
					for (int i=1; i<=numSamples; ++i)
					{
						Vector3 current = EvalBezierQuadraticPosition(i / numLines);
						distance += Vector3.Distance(previous, current);
						ourTempDistances[i] = distance;
						previous = current;
					}
				}
				else
				{
					for (int i=1; i<=numSamples; ++i)
					{
						Vector3 current = EvalBezierCubicPosition(i / numLines);
						distance += Vector3.Distance(previous, current);
						ourTempDistances[i] = distance;
						previous = current;
					}
				}

				distance += Vector3.Distance(previous, myTransition[myEntryCount-1].myEntry.myPosition);
				ourTempDistances[numSamples+1] = distance;

				t *= distance;

				int idx = Array.BinarySearch(ourTempDistances, 0, numSamples+2, t);
				if (idx < 0)
				{
					idx = ~idx;
					if (idx == 0)
						return 0.0f;
					else if (idx >= numSamples+2)
						return 1.0f;
					t = Mathf.InverseLerp(ourTempDistances[idx-1], ourTempDistances[idx], t);
					return Mathf.LerpUnclamped((idx-1) / numLines, idx / numLines, t);
				}
				else
				{
					return idx / numLines;
				}
			}

			private Vector3 EvalBezierLinearPosition(float t)
			{
				return Vector3.LerpUnclamped(myTransition[0].myEntry.myPosition, myTransition[1].myEntry.myPosition, t);
			}

			private Vector3 EvalBezierQuadraticPosition(float t)
			{
				// evaluating quadratic Bézier curve using Bernstein polynomials
				float s = 1.0f - t;
				return      (s*s) * myTransition[0].myEntry.myPosition
					 + (2.0f*s*t) * myTransition[1].myEntry.myPosition
					 +      (t*t) * myTransition[2].myEntry.myPosition;
			}

			private Vector3 EvalBezierCubicPosition(float t)
			{
				// evaluating cubic Bézier curve using Bernstein polynomials
				float s = 1.0f - t;
				float t2 = t*t;
				float s2 = s*s;
				return      (s*s2) * myTransition[0].myEntry.myPosition
					 + (3.0f*s2*t) * myTransition[1].myEntry.myPosition
					 + (3.0f*s*t2) * myTransition[2].myEntry.myPosition
					 +      (t*t2) * myTransition[3].myEntry.myPosition;
			}

			private Quaternion EvalBezierLinearRotation(float t)
			{
				return Quaternion.SlerpUnclamped(myTransition[0].myEntry.myRotation, myTransition[1].myEntry.myRotation, t);
			}

			private Quaternion EvalBezierQuadraticRotation(float t)
			{
				// evaluating quadratic Bézier curve using de Casteljau's algorithm
				ourTempQuaternions[0] = Quaternion.SlerpUnclamped(myTransition[0].myEntry.myRotation, myTransition[1].myEntry.myRotation, t);
				ourTempQuaternions[1] = Quaternion.SlerpUnclamped(myTransition[1].myEntry.myRotation, myTransition[2].myEntry.myRotation, t);
				return Quaternion.SlerpUnclamped(ourTempQuaternions[0], ourTempQuaternions[1], t);
			}

			private Quaternion EvalBezierCubicRotation(float t)
			{
				// evaluating cubic Bézier curve using de Casteljau's algorithm
				for (int i=0; i<3; ++i)
					ourTempQuaternions[i] = Quaternion.SlerpUnclamped(myTransition[i].myEntry.myRotation, myTransition[i+1].myEntry.myRotation, t);
				for (int i=0; i<2; ++i)
					ourTempQuaternions[i] = Quaternion.SlerpUnclamped(ourTempQuaternions[i], ourTempQuaternions[i+1], t);
				return Quaternion.SlerpUnclamped(ourTempQuaternions[0], ourTempQuaternions[1], t);
			}

			public void UpdateState(State state)
			{
				ControlEntryAnchored entry;
				if (state.myControlEntries.TryGetValue(this, out entry))
				{
					entry.Update();
					if (myApplyPosition)
						myTransform.position = entry.myEntry.myPosition;
					if (myApplyRotation)
						myTransform.rotation = entry.myEntry.myRotation;
				}
			}

			public bool IsValid()
			{
				return myTransform != null;
			}
		}

		private struct ControlEntry
		{
			public Quaternion myRotation;
			public Vector3 myPosition;
		}

		private class ControlEntryAnchored
		{
			public const int ANCHORMODE_WORLD = 0;
			public const int ANCHORMODE_SINGLE = 1;
			public const int ANCHORMODE_BLEND = 2;

			public ControlEntry myEntry;
			public ControlEntry myAnchorOffset;
			public Transform myAnchorATransform;
			public Transform myAnchorBTransform;
			public int myAnchorMode = ANCHORMODE_SINGLE;
			public float myBlendRatio = DEFAULT_ANCHOR_BLEND_RATIO;
			public float myDampingTime = DEFAULT_ANCHOR_DAMPING_TIME;

			public string myAnchorAAtom;
			public string myAnchorBAtom;
			public string myAnchorAControl = "control";
			public string myAnchorBControl = "control";
			
			public ControlEntryAnchored(IdlePoser plugin, string control)
			{
				Atom containingAtom = plugin.GetContainingAtom();
				if (plugin.myOptionsDefaultToWorldAnchor.val || containingAtom.type != "Person" || control == "control")
					myAnchorMode = ANCHORMODE_WORLD;
				myAnchorAAtom = myAnchorBAtom = containingAtom.uid;
			}
			
			public ControlEntryAnchored Clone()
			{			
				return (ControlEntryAnchored)MemberwiseClone();
			}

			public void Initialize()
			{
				GetTransforms();
				UpdateInstant();
			}

			public void AdjustAnchor()
			{
				GetTransforms();
				Capture(myEntry.myPosition, myEntry.myRotation);				
			}

			private void GetTransforms()
			{
				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myAnchorATransform = null;
					myAnchorBTransform = null;
				}
				else
				{
					myAnchorATransform = GetTransform(myAnchorAAtom, myAnchorAControl);
					if (myAnchorMode == ANCHORMODE_BLEND)
						myAnchorBTransform = GetTransform(myAnchorBAtom, myAnchorBControl);
					else
						myAnchorBTransform = null;
				}
			}

			private Transform GetTransform(string atomName, string controlName)
			{
				Atom atom = SuperController.singleton.GetAtomByUid(atomName);
				return atom?.GetStorableByID(controlName)?.transform;
			}

			public void UpdateInstant()
			{
				float dampingTime = myDampingTime;
				myDampingTime = 0.0f;
				Update();
				myDampingTime = dampingTime;
			}

			public void Update()
			{
				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myEntry = myAnchorOffset;
				}
				else
				{
					ControlEntry anchor;
					if (myAnchorMode == ANCHORMODE_SINGLE)
					{
						if (myAnchorATransform == null)
							return;
						anchor.myPosition = myAnchorATransform.position;
						anchor.myRotation = myAnchorATransform.rotation;
					}
					else
					{
						if (myAnchorATransform == null || myAnchorBTransform == null)
							return;
						anchor.myPosition = Vector3.LerpUnclamped(myAnchorATransform.position, myAnchorBTransform.position, myBlendRatio);
						anchor.myRotation = Quaternion.SlerpUnclamped(myAnchorATransform.rotation, myAnchorBTransform.rotation, myBlendRatio);
					}
					anchor.myPosition = anchor.myPosition + anchor.myRotation * myAnchorOffset.myPosition;
					anchor.myRotation = anchor.myRotation * myAnchorOffset.myRotation;

					if (myDampingTime >= 0.001f)
					{
						float t = Mathf.Clamp01(Time.deltaTime / myDampingTime);
						myEntry.myPosition = Vector3.LerpUnclamped(myEntry.myPosition, anchor.myPosition, t);
						myEntry.myRotation = Quaternion.SlerpUnclamped(myEntry.myRotation, anchor.myRotation, t);
					}
					else
					{
						myEntry = anchor;
					}
				}
			}

			public void Capture(Vector3 position, Quaternion rotation)
			{
				myEntry.myPosition = position;
				myEntry.myRotation = rotation;

				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myAnchorOffset.myPosition = position;
					myAnchorOffset.myRotation = rotation;
				}
				else
				{
					ControlEntry root;
					if (myAnchorMode == ANCHORMODE_SINGLE)
					{
						if (myAnchorATransform == null)
							return;
						root.myPosition = myAnchorATransform.position;
						root.myRotation = myAnchorATransform.rotation;
					}
					else
					{
						if (myAnchorATransform == null || myAnchorBTransform == null)
							return;
						root.myPosition = Vector3.LerpUnclamped(myAnchorATransform.position, myAnchorBTransform.position, myBlendRatio);
						root.myRotation = Quaternion.SlerpUnclamped(myAnchorATransform.rotation, myAnchorBTransform.rotation, myBlendRatio);
					}

					myAnchorOffset.myPosition = Quaternion.Inverse(root.myRotation) * (position - root.myPosition);
					myAnchorOffset.myRotation = Quaternion.Inverse(root.myRotation) * rotation;
				}
			}
		}


		private class MorphCapture
		{
			public string mySID;
			public DAZMorph myMorph;
			public DAZCharacterSelector.Gender myGender;
			private float[] myTransition = new float[MAX_STATES];
			private int myEntryCount = 0;
			public bool myApply = true;

			// used when adding a capture
			public MorphCapture(IdlePoser plugin, DAZCharacterSelector.Gender gender, DAZMorph morph)
			{
				myMorph = morph;
				myGender = gender;
				mySID = plugin.GenerateMorphsSID(gender == DAZCharacterSelector.Gender.Female);
			}

			// legacy handling of old qualifiedName where the order was reversed
			public MorphCapture(IdlePoser plugin, DAZCharacterSelector geometry, string oldQualifiedName)
			{
				
				bool isFemale = oldQualifiedName.StartsWith("Female#");
				if (!isFemale && !oldQualifiedName.StartsWith("Male#"))
					return;
				GenerateDAZMorphsControlUI morphsControl = isFemale ? geometry.morphsControlFemaleUI : geometry.morphsControlMaleUI;
				string morphUID = oldQualifiedName.Substring(isFemale ? 7 : 5);
				myMorph = morphsControl.GetMorphByUid(morphUID);
				myGender = isFemale ? DAZCharacterSelector.Gender.Female : DAZCharacterSelector.Gender.Male;

				mySID = plugin.GenerateMorphsSID(isFemale);
			}
			
			// legacy handling before there were ShortIDs
			public MorphCapture(IdlePoser plugin, DAZCharacterSelector geometry, string morphUID, bool isFemale)
			{
				
				GenerateDAZMorphsControlUI morphsControl = isFemale ? geometry.morphsControlFemaleUI : geometry.morphsControlMaleUI;
				myMorph = morphsControl.GetMorphByUid(morphUID);
				myGender = isFemale ? DAZCharacterSelector.Gender.Female : DAZCharacterSelector.Gender.Male;

				mySID = plugin.GenerateMorphsSID(isFemale);
			}
			
			// used when loading from JSON
			public MorphCapture(DAZCharacterSelector geometry, string morphUID, string morphSID)
			{
				bool isFemale = morphSID.Length > 0 && morphSID[0] == 'F';
				GenerateDAZMorphsControlUI morphsControl = isFemale ? geometry.morphsControlFemaleUI : geometry.morphsControlMaleUI;
				myMorph = morphsControl.GetMorphByUid(morphUID);
				myGender = isFemale ? DAZCharacterSelector.Gender.Female : DAZCharacterSelector.Gender.Male;
				mySID = morphSID;
			}

			public void CaptureEntry(State state)
			{
				state.myMorphEntries[this] = myMorph.morphValue;
			}

			public void SetTransition(List<State> states, int entryCount)
			{
				myEntryCount = entryCount;
				bool identical = true;
				float morphValue = myMorph.morphValue;
				for (int i=0; i<myEntryCount; ++i)
				{
					if (!states[i].myMorphEntries.TryGetValue(this, out myTransition[i]))
					{
						CaptureEntry(states[i]);
						myTransition[i] = morphValue;
					}
					else
					{
						identical &= (myTransition[i] == morphValue);
					}
				}
				if (identical)
					myEntryCount = 0; // nothing to do, save some performance
			}

			public void UpdateTransition(float t)
			{
				if (!myApply)
					return;

				switch (myEntryCount)
				{
					case 4:
						myMorph.morphValue = EvalBezierCubic(t);
						break;
					case 3:
						myMorph.morphValue = EvalBezierQuadratic(t);
						break;
					case 2:
						myMorph.morphValue = EvalBezierLinear(t);
						break;
					default:
						myMorph.morphValue = myTransition[0];
						break;
				}
			}

			private float EvalBezierLinear(float t)
			{
				return Mathf.LerpUnclamped(myTransition[0], myTransition[1], t);
			}

			private float EvalBezierQuadratic(float t)
			{
				// evaluating using Bernstein polynomials
				float s = 1.0f - t;
				return      (s*s) * myTransition[0]
					 + (2.0f*s*t) * myTransition[1]
					 +      (t*t) * myTransition[2];
			}

			private float EvalBezierCubic(float t)
			{
				// evaluating using Bernstein polynomials
				float s = 1.0f - t;
				float t2 = t*t;
				float s2 = s*s;
				return      (s*s2) * myTransition[0]
					 + (3.0f*s2*t) * myTransition[1]
					 + (3.0f*s*t2) * myTransition[2]
					 +      (t*t2) * myTransition[3];
			}

			public bool IsValid()
			{
				return myMorph != null && mySID != null;
			}
		}
		
		private string GenerateMorphsSID(bool isFemale)
		{
			string chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
			char[] randomChars = new char[6];
			randomChars[0] = isFemale ? 'F' : 'M';
			randomChars[1] = '-';
			for (int a=0; a<10; ++a)
			{
				// find unused shortID
				for (int i=2; i<randomChars.Length; ++i)
					randomChars[i] = chars[UnityEngine.Random.Range(0, chars.Length)];
				string sid = new string(randomChars);
				if (myMorphCaptures.Find(x => x.mySID == sid) == null)
					return sid;
			}
			
			return null; // you are very lucky, you should play lottery!
		}
	}
}